# Back-Propagation Neural Networks
# 
# Written in Python.  See http://www.python.org/
# Placed in the public domain.
# Neil Schemenauer <nas@arctrix.com>

import math
import random
import string

random.seed(0)

# calculate a random number where:  a <= rand < b
def rand(a, b):
    return (b-a)*random.random() + a

# Make a matrix (we could use NumPy to speed this up)
def makeMatrix(I, J, fill=0.0):
    m = []
    for i in range(I):
        m.append([fill]*J)
    return m

# our sigmoid function, tanh is a little nicer than the standard 1/(1+e^-x)
# tanh gir verdier fra -1 til 1, sigmoid gir fra 0 til 1 som passer mye bedre i mitt tilfelle
def sigmoid(x):
    return 1/(1+math.exp(-x))
    #return math.tanh(x)

# derivative of our sigmoid function
def dsigmoid(y):
    return y * (1 - y)
    #return 1.0-y*y

#sigmoid <- function (x)  1 / (1 + exp(-x))
#dsigmoid <- function (x)  x * (1 - x)



class NN:
    def __init__(self, ni, nh, no):
        # number of input, hidden, and output nodes
        self.ni = ni + 1 # +1 for bias node
        self.nh = nh
        self.no = no

        # activations for nodes
        self.ai = [1.0]*self.ni
        self.ah = [1.0]*self.nh
        self.ao = [1.0]*self.no
        
        # create weights
        self.wi = makeMatrix(self.ni, self.nh)
        self.wo = makeMatrix(self.nh, self.no)
        # set them to random vaules
        for i in range(self.ni):
            for j in range(self.nh):
                self.wi[i][j] = rand(-2.0, 2.0)
        for j in range(self.nh):
            for k in range(self.no):
                self.wo[j][k] = rand(-2.0, 2.0)

        # last change in weights for momentum   
        self.ci = makeMatrix(self.ni, self.nh)
        self.co = makeMatrix(self.nh, self.no)
        print 'Opprettet nettverk (',ni,nh,no,')'

    def update(self, inputs):
        if len(inputs) != self.ni-1:
            raise ValueError, 'wrong number of inputs'

        # input activations
        for i in range(self.ni-1):
            #self.ai[i] = sigmoid(inputs[i])
            self.ai[i] = inputs[i]

        # hidden activations
        for j in range(self.nh):
            sum = 0.0
            for i in range(self.ni):
                sum = sum + self.ai[i] * self.wi[i][j]
            self.ah[j] = sigmoid(sum)

        # output activations
        for k in range(self.no):
            sum = 0.0
            for j in range(self.nh):
                sum = sum + self.ah[j] * self.wo[j][k]
            self.ao[k] = sigmoid(sum)

        return self.ao[:]


    def backPropagate(self, targets, N, M):
        if len(targets) != self.no:
            raise ValueError, 'wrong number of target values'

        # calculate error terms for output
        output_deltas = [0.0] * self.no
        for k in range(self.no):
            error = targets[k]-self.ao[k]
            output_deltas[k] = dsigmoid(self.ao[k]) * error

        # calculate error terms for hidden
        hidden_deltas = [0.0] * self.nh
        for j in range(self.nh):
            error = 0.0
            for k in range(self.no):
                error = error + output_deltas[k]*self.wo[j][k]
            hidden_deltas[j] = dsigmoid(self.ah[j]) * error

        # update output weights
        for j in range(self.nh):
            for k in range(self.no):
                change = output_deltas[k]*self.ah[j]
                self.wo[j][k] = self.wo[j][k] + N*change + M*self.co[j][k]
                self.co[j][k] = change
                #print N*change, M*self.co[j][k]

        # update input weights
        for i in range(self.ni):
            for j in range(self.nh):
                change = hidden_deltas[j]*self.ai[i]
                self.wi[i][j] = self.wi[i][j] + N*change + M*self.ci[i][j]
                self.ci[i][j] = change

        # calculate error
        error = 0.0
        for k in range(len(targets)):
            error = error + 0.5*(targets[k]-self.ao[k])**2
        return error


    def test(self, patterns):
        for p in patterns:
            print p[0], '->', self.update(p[0])

    def weights(self):
        print 'Input weights:'
        for i in range(self.ni):
            print self.wi[i]
        print
        print 'Output weights:'
        for j in range(self.nh):
            print self.wo[j]

    def train(self, patterns, iterations=1000, N=0.5, M=0.1, stoperr=0.0):
        """Trener opp nettverket."""
        # N: learning rate
        # M: momentum factor
        minerr = 100
        minerri = 0
        for i in xrange(iterations):
            error = 0.0
            for p in patterns:
                inputs = p[0]
                targets = p[1]
                self.update(inputs)
                error = error + self.backPropagate(targets, N, M)
            if i % (int(iterations/10)) == 0:
                print i,'error %-14f' % error, 'min:',minerri, minerr
            if error<minerr:
                minerr=error
                minerri=i
            if i==(iterations-1):
                print 'Ferdig. Error %-14f' % error
                print 'Min error:',minerr,'index:',minerri
            if error<stoperr:
                print 'Ferdig ved < stoperr. Error %-14f' % error
                print 'Min error:',minerr,'index:',minerri
                break
        
    def finnMinErrIt(self, patterns, maxit=1000):
        """Minner beste antall iterasjoner for treningssett."""
        iterations=maxit
        N=0.5
        M=0.1
        print 'Leter etter index for minste error'
        minerr = 100
        minerri = 0
        for i in xrange(iterations):
            error = 0.0
            for p in patterns:
                inputs = p[0]
                targets = p[1]
                self.update(inputs)
                error = error + self.backPropagate(targets, N, M)
            if i % 100 == 0:
                print i,'error %-14f' % error
            if error<minerr:
                minerr=error
                minerri=i
            if i==(iterations-1):
                print 'Ferdig. Error %-14f' % error
                print 'Min error:',minerr,'index:',minerri
        return minerri

    def finn(self, patterns):
        for p in patterns:
            ret = self.update(p[0])
            #print p[0], '->', int(round(ret[0])),int(round(ret[1])),int(round(ret[2]))
            print '(',p[1],')->', round(ret[0],2),round(ret[1],2),round(ret[2],2),round(ret[3],2)

    def finnVal(self, pattern):
        ret = self.update(pattern)
        return [int(round(ret[0],0)), int(round(ret[1],0)), int(round(ret[2],0)), int(round(ret[3],0))]

    def finnValError(self, pattern):
        ret = self.update(pattern)
        e=0
        for r in ret:
            if r>0.5:
                e+=1-r
            else:
                e+=r
        print round(ret[0],1), round(ret[1],1), round(ret[2],1), round(ret[3],1), 'Total error:', round(e,2)

    def finnValE(self, pattern):
        ret = self.update(pattern)
        e=0
        for r in ret:
            if r>0.5:
                e+=1-r
            else:
                e+=r
        return [[int(round(ret[0],0)), int(round(ret[1],0)), int(round(ret[2],0)), int(round(ret[3],0))],e]

def test():
    ''' Akkorder (1-7:dur,:8-14:moll, 0&15:ingen akk)
    1-C 2-D 3-E 4-F 5-G 6-A 7-H
    
    
    '''
    pat = [
        # 1 2 3 4 5 6 7 8 9 101112
        [[0,0,0,0,0,0,0,0,0,0,0,0], [0,0,0,0]],
        [[1,0,0,0,1,0,0,1,0,0,0,0], [0,0,0,1]], # C
        [[0,0,1,0,0,0,1,0,0,1,0,0], [0,0,1,0]], # D
        [[0,0,0,0,1,0,0,0,1,0,0,1], [0,0,1,1]], # E
        [[1,0,0,0,0,1,0,0,0,1,0,0], [0,1,0,0]], # F
        [[0,0,1,0,0,0,0,1,0,0,0,1], [0,1,0,1]], # G
        [[0,1,0,0,1,0,0,0,0,1,0,0], [0,1,1,0]], # A
        [[0,0,0,1,0,0,1,0,0,0,0,1], [0,1,1,1]], # H
        [[1,0,1,1,1,1,0,1,0,1,0,0], [0,0,0,0]], # --
        [[1,0,1,0,1,0,0,1,1,1,0,0], [0,0,0,0]], # --
        [[1,0,1,1,1,0,1,1,0,1,0,0], [0,0,0,0]], # --
        [[1,0,1,0,1,0,1,0,1,1,0,0], [0,0,0,0]], # --
        [[1,0,1,1,1,0,0,1,1,1,0,0], [0,0,0,0]], # --
        [[1,0,1,0,1,0,0,1,1,1,1,0], [0,0,0,0]], # --
        [[1,0,1,1,1,0,0,1,0,1,0,0], [0,0,0,0]], # --
        [[1,0,1,0,1,0,0,1,1,1,1,1], [0,0,0,0]], # --
        [[1,1,1,1,1,0,0,1,0,1,0,0], [0,0,0,0]], # --
        [[1,0,0,1,1,0,0,1,1,1,0,0], [0,0,0,0]], # --
        [[1,0,1,1,0,0,0,0,0,0,0,0], [0,0,0,0]], # --
        [[1,0,1,0,1,0,0,1,1,1,0,0], [0,0,0,0]], # --
        [[1,1,0,0,0,0,0,1,1,1,0,0], [0,0,0,0]], # --
        [[0,0,1,1,1,0,0,0,0,0,0,0], [0,0,0,0]], # --
    ]
    
    testpat = [
        [[0,0,0,0,0,0,0,0,0,0,0,0], [0,0,0,0]],
        [[1,0,0,0,1,0,0,1,0,1,0,0], [0,0,0,1]], # C
        [[0,0,1,0,0,0,1,0,0,1,0,0], [0,0,1,0]], # D
        [[0,0,0,0,1,0,0,0,1,0,0,1], [0,0,1,1]], # E
        [[1,0,0,0,0,1,0,0,0,1,0,0], [0,1,0,0]], # F
        [[0,0,1,0,0,0,0,1,0,0,0,1], [0,1,0,1]], # G
        [[1,1,1,1,1,0,0,1,0,1,0,0], [0,0,0,0]], # --
        [[1,0,0,1,1,0,0,1,1,1,1,0], [0,0,0,0]], # --
        [[1,0,1,1,0,0,0,0,0,0,0,0], [0,0,0,0]], # --
        [[1,0,1,0,1,0,0,0,0,1,0,0], [0,0,0,0]], # --
        [[0,1,0,0,0,0,0,1,0,0,0,0], [0,0,0,0]], # --
        [[0,0,1,0,1,0,0,0,0,0,1,1], [0,0,0,0]], # --
    ]

    # create a network with two input, two hidden, and one output nodes
    n = NN(12, 4, 4)
    # train it with some patterns
    n.train(pat)
    # test it
    #n.finn(testpat)
    print n.finnVal([1,1,1,1,1,1,0,0,0,0,1,1])



if __name__ == '__main__':
    test()
